from colorama import Fore, Style
from werkzeug.utils import send_from_directory
from .validation import validate
from .route_management import Route
from werkzeug import Request, Response
from marsrouter import Router
import importlib.util
import json
import os

class WSGIApp:
	def __init__(self):
		self.routes = []
		self.load_project()
		self.router = Router()
		self.load_mapper()

	def load_mapper(self):
		for route in self.routes:
			if route.url[-1] == "/" and route.url != "/":
				route.url = route.url[:-1]
			self.router.add_route(route.url, route.handler, methods=[route.method])

	def load_project(self):
		# Load the map.json file
		with open('map.json') as f:
			config = json.load(f)
		
		# Load and register routes from controllers
		for controller_path in set(config.get('controllers', [])):
			self._load_controller(controller_path)

	def _load_controller(self, controller_path):
		# Import the module
		module_name = os.path.splitext(os.path.basename(controller_path))[0]
		spec = importlib.util.spec_from_file_location(module_name, controller_path)
		module = importlib.util.module_from_spec(spec)
		spec.loader.exec_module(module)
		
		# Get the controller class name (assuming it's the same as the module name)
		controller_class_name = module_name 
		
		# Get the controller class from the module
		controller_class = getattr(module, controller_class_name, None)
		if not controller_class:
			print(f"{Fore.YELLOW}[WARNING] Controller class {controller_class_name} not found in module {module_name}{Style.RESET_ALL}")
			return
	
		# Iterate over all attributes in the class
		for method_name in dir(controller_class):
			method = getattr(controller_class, method_name)
			
			# Check if the attribute is callable and has route attributes
			if callable(method) and hasattr(method, '_route_method') and hasattr(method, '_route_url'):
				route = Route(method._route_method, method._route_url, method)
				self.routes.append(route)

	def wsgi_handler(self, environ, start_response):
		# Create a Request object from WSGI environment
		request = Request(environ)
		
		method = request.method
		path = request.path
		if path.startswith('/static/'):
			# Serve the static file from the directory
			filename = path[len('/static/'):]
			try:
				response = send_from_directory("src/static", filename, environ, as_attachment=False)
			except:
				response = Response("<h1>404 Not Found</h1>", status=404, content_type='text/html')
				return response(environ, start_response)
		
		result = self.router.match(path, method)
		if not result.get("controller"):
			return Response(result.get("error"), status=result.get("status_code"), content_type='text/html')(environ, start_response)
		controller = result.get("controller")
		if not callable(controller):
			raise ValueError(f"Controller '{controller}' is not a function")
		args = result.get("params")
		if not isinstance(args, dict):
			raise ValueError("Params must be a dict")

		if hasattr(controller, "_auth"):
			authorization = controller._auth
			cookie = controller._auth_cookie
			error = controller._error
			negative_auth = controller._negative_auth
			token = None

			if cookie and not negative_auth: #Only raise an error if negative auth is not present 
				token = request.cookies.get(cookie)
				if not token:
					if error:
						return error()(environ, start_response)
					rsp = Response(f"Error: No JWT found in the '{cookie}' cookie")
					return rsp(environ, start_response)
			elif not negative_auth: #Only raise an error if negative auth is not present
				token = request.headers.get("Authorization")
				if not token:
					if error:
						return error()(environ, start_response)
					rsp = Response("Error: No JWT token found in the Autherization header")
					rsp.status_code = 400
					return rsp(environ, start_response)
				if token.startswith("Bearer"):
					token = token[7:]

			if not negative_auth and not authorization._verify(token): #Only raise an error if negative auth is not present
				if error:
					return error()(environ, start_response)
				rsp = Response("Error: Invalid signature in token")
				rsp.status_code = 403
				return rsp(environ, start_response)
			#If you passed all the stages of verification, and you have negative auth, raise an error
			if negative_auth:
				if error:
					return error()(environ, start_response)
				rsp = Response("Error: The token is valid, this route requires the token to be invalid")
				rsp.status_code = 400
				return rsp(environ, start_response)

		if hasattr(controller, "_validator"):
			validator = controller._validator
			error = controller._error
			mimetypes = controller._mimetypes
			if mimetypes and not request.mimetype in mimetypes:
				if error:
					return error()(environ, start_response)
				rsp = Response("Error: Requested content type is not supported")
				rsp.status_code = 400
				return rsp(environ, start_response)
			# Go through the request data, only json and html are supported
			try:
				data = request.json
			except:
				try:
					data = request.form
				except:
					if error:
						return error()
					rsp = Response("Error: No data provided")
					rsp.status_code = 400
					return rsp
			if not data:
				if error:
					return error()
				rsp = Response("Error: No data provided or data was malformed")
				rsp.status_code = 400
				return rsp(environ, start_response)

			validation_result = validate(validator, error, data)
			if validation_result:
				return validation_result(environ, start_response)

		return controller(request, *args.values())(environ, start_response)

	def __call__(self, environ, start_response):
		return self.wsgi_handler(environ, start_response)




from functools import wraps
from .validate import validate
import re
class String:
	def __init__(self, **kwargs) -> None:
		self.kwargs = kwargs
		self.min = self.kwargs.get("min")
		self.max = self.kwargs.get("max")
		
		self.minimum = self.kwargs.get("minimum")
		self.maximum = self.kwargs.get("maximum")

		self.regex = self.kwargs.get("regex")

		if self.min == None:
			self.min = self.minimum
		if self.max == None:
			self.max = self.maximum

	def validate(self, string) -> bool:
		if type(string) != str:
			return False
		if self.regex:
			# Using fullmatch to ensure the whole string matches the pattern
			if re.compile(self.regex).fullmatch(string):
				return True
			else:
				return False

		if self.min:
			if len(string) < self.min:
				return False
		
		if self.max:
			if len(string) > self.max:
				return False

		return True #Return true if all conditions are met

class Integer:
	def __init__(self, **kwargs) -> None:
		self.kwargs = kwargs
		self.min = self.kwargs.get("min")
		self.max = self.kwargs.get("max")
		
		self.minimum = self.kwargs.get("minimum")
		self.maximum = self.kwargs.get("maximum")

		if self.min == None:
			self.min = self.minimum
		if self.max == None:
			self.max = self.maximum

	def validate(self, integer) -> bool:
		if type(integer) != int:
			return False

		if self.min:
			if integer < self.min:
				return False
		
		if self.max:
			if integer > self.max:
				return False

		return True #Return true if all conditions are met

class Float:
	def __init__(self, **kwargs) -> None:
		self.kwargs = kwargs
		self.min = self.kwargs.get("min")
		self.max = self.kwargs.get("max")
		
		self.minimum = self.kwargs.get("minimum")
		self.maximum = self.kwargs.get("maximum")

		if self.min is None:
			self.min = self.minimum
		if self.max is None:
			self.max = self.maximum

	def validate(self, value) -> bool:
		if type(value) not in [float, int]:  # Allow int as it can be considered a valid float
			return False

		if self.min is not None:
			if value < self.min:
				return False
		
		if self.max is not None:
			if value > self.max:
				return False

		return True  # Return true if all conditions are met

class Array:
	def __init__(self, element_type=None, **kwargs) -> None:
		self.element_type = element_type
		self.kwargs = kwargs
		self.min = self.kwargs.get("min")
		self.max = self.kwargs.get("max")

		self.minimum = self.kwargs.get("minimum")
		self.maximum = self.kwargs.get("maximum")

		if self.min is None:
			self.min = self.minimum
		if self.max is None:
			self.max = self.maximum

	def validate(self, array) -> bool:
		if not isinstance(array, list):  # Ensure the input is a list
			return False

		if self.min is not None:
			if len(array) < self.min:
				return False

		if self.max is not None:
			if len(array) > self.max:
				return False

		if self.element_type:  # If element_type is provided, validate each element
			for item in array:
				if hasattr(self.element_type, "__nested_validator"):
					# If the element_type is an Object, validate the nested structure
					if not isinstance(item, dict) or validate(self.element_type, None, item):
						return False
				elif not self.element_type.validate(item):
					return False

		return True  # Return true if all conditions are met

class Boolean:
	def __init__(self) -> None:
		pass #Its a boolean, what do you want it to do
			
	def validate(self, boolean) -> bool:
		if type(boolean) != boolean:
			return False
		return True

class Union:
	def __init__(self, *args) -> None:
		self.types = args

	def validate(self, any) -> bool:
		for typechecks in self.types:
			if typechecks.validate(any):
				#If a type check passes, it is valid
				return True
		return False

class Any:
	def validate(self, value) -> bool:
		return True

def Object(**kwargs):
	exclude_from_validation = kwargs.get('exclude_from_validation', False)
	
	def decorator(func):
		@wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, **kwargs)
		wrapper.__nested_validator = True
		wrapper.__exclude_from_validation = exclude_from_validation
		return wrapper
	return decorator




from sqlalchemy.ext.declarative import declarative_base
Base = declarative_base()
from sqlalchemy import ForeignKey, Column, ARRAY, BIGINT, BINARY, BLOB, BOOLEAN, BigInteger, Boolean, CHAR, CLOB, DATE, DATETIME, DECIMAL, Date, DateTime, Enum, FLOAT, Float, INT, INTEGER, Integer, Interval, JSON, LargeBinary, NCHAR, NUMERIC, NVARCHAR, Numeric, PickleType, REAL, SMALLINT, SmallInteger, String, TEXT, TIME, TIMESTAMP, Text, Time, TupleType, TypeDecorator, Unicode, UnicodeText, VARBINARY, VARCHAR 
from sqlalchemy.orm import relationship
from .connection import connection
from .MigrationGenerator import MigrationSystem, MigrationWrapper



from Crypto.Protocol.KDF import PBKDF2
from Crypto.Random import get_random_bytes
from Crypto.Hash import SHA256
import base64
def generate_salt() -> bytes:
	"""Generate a random salt."""
	return get_random_bytes(16)

# Function to hash a password with a given salt
def hash_password(password: str, salt: bytes) -> str:
	"""Hash a password with a given salt using PBKDF2."""
	key = PBKDF2(password, salt, dkLen=32, count=100000, hmac_hash_module=SHA256)
	return base64.b64encode(key).decode('utf-8')

# Function to verify a password against the stored hash
def verify_password(stored_password: str, password: str, salt: bytes) -> bool:
	"""Verify a password against the stored hash."""
	hashed_password = hash_password(password, salt)
	return stored_password == hashed_password



from json import loads, dumps
from colorama import Fore, Style
from libmercury.db.setup_db import * 
from libmercury.security import keygen 
from libmercury.db import MigrationSystem
from libmercury.generation import generate
from .version import version
import os
import importlib.util
class CLI:
	def __init__(self, arguments) -> None:
		self.arguments = arguments

	def execute(self):
		try:
			with open(".mercury", "r") as f:
				os.chdir(f.read())
		except:
			pass
		del self.arguments[0]
		commands = {
			"init": self.init,
			"create": self.create,
			"migrate": self.migrate,
			"generate": self.generate,
			"run": self.run,
		}
		if len(self.arguments) < 1:
			self.version_display()
			return
		try:
			commands[self.arguments[0]]()
		except KeyError:
			self.unknown_command()

	def init(self):
		print(f"{Fore.BLUE}[Initializer]{Style.RESET_ALL} Launching Mercury {version} initializer...")
		directory = input(f"{Fore.BLUE}[Initializer]{Style.RESET_ALL} Name of directory to use: ")
		interpreter = input(f"{Fore.BLUE}[Initializer]{Style.RESET_ALL} Please provide your python interpreter(python, python3, etc): ")

		#Create project file structure
		os.mkdir(directory) 
		os.mkdir(f"{directory}/src")
		os.mkdir(f"{directory}/src/templates")
		os.mkdir(f"{directory}/src/static")
		os.mkdir(f"{directory}/src/controllers")
		os.mkdir(f"{directory}/src/validators")
		os.mkdir(f"{directory}/src/cargo")
		os.mkdir(f"{directory}/src/cargo/migrations")
		os.mkdir(f"{directory}/src/security")
		os.mkdir(f"{directory}/src/.vault")
		
		#Create files
		with open(f"{directory}/map.json", "w") as f:
			f.write(dumps({
				"interpreter": interpreter,
				"version": version,
				"controllers": [],
				"models": [],
				"validators": [],
				"security": []
			}))

		with open(f"{directory}/src/cargo/dev.db", "w") as f:
			f.write("")

		with open(f"{directory}/src/cargo/connection.py", "w") as f:
			f.write("""from libmercury.db import connection
Connection = connection("sqlite:///src/cargo/dev.db", echo=False)
#Connection.Engine - The engine
#Connection.Session - The session connected to the db""")
		with open(f"{directory}/app.py", "w") as f:
			f.write("""from libmercury.wsgi import WSGIApp
from werkzeug.serving import run_simple
app = WSGIApp()
run_simple("localhost", 8000, app)""")

		#Create .mercury files that allow us to run commands from anywhere in the file structure
		directories = [
			"src",
			"src/controllers",
			"src/validators",
			"src/security",
			"src/cargo",
			"src/cargo/migrations",
			"src/templates",
			"src/static"
		]
		for i in directories:
			with open(f"{directory}/{i}/.mercury", "w") as f:
				f.write(f"{os.getcwd()}/{directory}")
		
		create_mercury_table(f"sqlite:///{directory}/src/cargo/dev.db")
		print(f"{Fore.BLUE}[Initializer]{Style.RESET_ALL} Initialized project in directory: '{directory}'")

	def _import_module(self, file_path):
		module_name = os.path.splitext(os.path.basename(file_path))[0]
		spec = importlib.util.spec_from_file_location(module_name, file_path)
		module = importlib.util.module_from_spec(spec)
		spec.loader.exec_module(module)
		return module

	def create(self):
		if len(self.arguments) < 3:
			print(f"{Fore.RED}Error:{Style.RESET_ALL} Command 'create' requires at least 2 parameters")
			print("Usage:")
			print("create <thing> <named>")
			return
		
		thing = self.arguments[1]
		named = self.arguments[2]

		things = {
			"controller": self._create_controller,
			"validator": self._create_validator,
			"model": self._create_model,
			"jwt": self._create_jwt,
			"migration": self._create_migration
		}
		try:
			things[thing](named)
		except KeyError as e:
			print(f"{Fore.RED}Error:{Style.RESET_ALL} Unknown createable {thing}")
			print("Usage:")
			print("create <thing> <named>")
	
	def _create_controller(self, name):
		#Create placeholder
		with open(f"src/controllers/{name}Controller.py", "w") as f:
			f.write(f"""from libmercury import GETRoute, Request, Response
class {name}Controller:
	@staticmethod
	@GETRoute("/example")
	def example(request: Request) -> Response:
		response = Response("<h1>Example Page</h1>")
		response.headers['Content-Type'] = 'text/html'
		return response""")
		
		#Update Map.json
		with open("map.json", "r") as f:
			map_json = loads(f.read())
			map_json["controllers"].append(f"src/controllers/{name}Controller.py")
			map_json["controllers"] = list(set(map_json["controllers"]))
			map_json["validators"] = list(set(map_json["validators"]))
			map_json["models"] = list(set(map_json["models"]))
			map_json["security"] = list(set(map_json["security"]))
		with open("map.json", "w") as f:
			f.write(dumps(map_json))

		print(f"{Fore.BLUE}[CODEGEN]{Style.RESET_ALL} Successfully created src/controllers/{name}Controller.py")

	def _create_validator(self, name):
		with open(f"src/validators/{name}Validator.py", "w") as f:
			f.write(f"""from libmercury import Validator 
class {name}Validator:
	pass
		""")
		
		#Update Map.json
		with open("map.json", "r") as f:
			map_json = loads(f.read())
			map_json["validators"].append(f"src/validators/{name}Validator.py")
			map_json["controllers"] = list(set(map_json["controllers"]))
			map_json["validators"] = list(set(map_json["validators"]))
			map_json["models"] = list(set(map_json["models"]))
			map_json["security"] = list(set(map_json["security"]))
		with open("map.json", "w") as f:
			f.write(dumps(map_json))

		print(f"{Fore.BLUE}[CODEGEN]{Style.RESET_ALL} Successfully created src/validators/{name}Validator.py")

	def _create_migration(self, message):
		#Get all models
		with open("map.json", "r") as f:
			map_json = loads(f.read())
			model_paths = map_json["models"]
			map_json["controllers"] = list(set(map_json["controllers"]))
			map_json["validators"] = list(set(map_json["validators"]))
			map_json["models"] = list(set(map_json["models"]))
			map_json["security"] = list(set(map_json["security"]))

		#Run migrator
		print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Starting Migrator")
		migrator = MigrationSystem("src/cargo/connection.py", model_paths)
		migrator._create_migration(message)

	def _create_model(self, name):
		with open(f"src/cargo/{name}Model.py", "w") as f:
			f.write(f"""from libmercury.db import Column, Integer, Base

class {name}(Base):
	__tablename__ = "{name}"
	id = Column(Integer, primary_key=True)""")
		with open("map.json", "r") as f:
			map_json = loads(f.read())
			map_json["models"].append(f"src/cargo/{name}Model.py")
			map_json["controllers"] = list(set(map_json["controllers"]))
			map_json["validators"] = list(set(map_json["validators"]))
			map_json["models"] = list(set(map_json["models"]))
			map_json["security"] = list(set(map_json["security"]))

		with open("map.json", "w") as f:
			f.write(dumps(map_json))
		print(f"{Fore.BLUE}[CODEGEN]{Style.RESET_ALL} Successfully created src/cargo/{name}Model.py")

	def _create_jwt(self, name):
		key_type = keygen.main(name)
		public_key = f"{name}Public_key.pem"
		private_key = f"{name}Private_key.pem"
		if key_type == "HMAC":
			public_key = f"{name}Hmac_secret.key"
			private_key = f"{name}Hmac_secret.key"
		with open(f"src/security/{name}Jwt.py", "w") as f:
			f.write(f"""from libmercury.security import JWT
class {name}Jwt:
	@staticmethod
	def _makeJwt(body:dict):
		jwt = JWT("")
		jwt.payload = body
		return jwt.sign("src/.vault/{private_key}", "{key_type}")

	@staticmethod
	def _verify(jwt):
		try:
			jwt = JWT(jwt)
			return jwt.verify_signature("src/.vault/{public_key}")
		except ValueError:
			return False""")
		with open("map.json", "r") as f:
			map_json = loads(f.read())
			map_json["security"].append(f"src/secuirty/{name}Jwt.py")
			map_json["controllers"] = list(set(map_json["controllers"]))
			map_json["validators"] = list(set(map_json["validators"]))
			map_json["models"] = list(set(map_json["models"]))
			map_json["security"] = list(set(map_json["security"]))

		with open("map.json", "w") as f:
			f.write(dumps(map_json))
		
		print(f"{Fore.BLUE}[CODEGEN]{Style.RESET_ALL} Successfully created src/cargo/{name}Jwt.py")

	def migrate(self):
		# Extract the database URL from the `Connection` object
		module = self._import_module("src/cargo/connection.py")
		if hasattr(module, 'Connection'):
			connection = module.Connection
			# Extracting the database URL from the connection object
			# This depends on the actual implementation of `connection` in `libmercury.db`
			if hasattr(connection, 'Engine'):
				db_url = connection.Engine.url
			else:
				raise AttributeError("The 'Connection' object does not have an 'engine' or 'url' attribute.")
		else:
			raise AttributeError("The module does not have a 'Connection' object.")

		if find_mercury_table(db_url) is None:
			create_mercury_table(db_url)
			create_version(db_url, 0)	
		db_version = get_version(db_url)
		latest_migration_id = 0
		migrations = []

		#Get all non-runned migrations
		for file in os.listdir("src/cargo/migrations"):
			if file.endswith('.py') and os.path.isfile(os.path.join("src/cargo/migrations", file)):
				try:
					if int(file[:-3]) > db_version:
						if int(file[:-3]) > latest_migration_id:
							latest_migration_id = int(file[:-3])
						migrations.append(os.path.join("src/cargo/migrations", file))
				except ValueError:
					pass
		migrations.sort(key=lambda x: int(os.path.basename(x)[:-3]))
		
		for migration in migrations:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Running migration {migration}")
			try:
				module = self._import_module(migration).upgrade(db_url)
				print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} '{migration}' passed with no errors")
			except Exception as e:
				print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Migration: '{migration}' failed with error:")
				print(e)

		#Update map
		map = loads(open("map.json", "r").read())
		if not latest_migration_id == 0:
			update_version(db_url, latest_migration_id)

		with open("map.json", "w") as f:
			map["controllers"] = list(set(map["controllers"]))
			map["validators"] = list(set(map["validators"]))
			map["models"] = list(set(map["models"]))
			map["security"] = list(set(map["security"]))
			f.write(dumps(map))

	def run(self):
		with open("map.json", "r") as f:
			map = loads(f.read())
		os.system(f"{map['interpreter']} app.py")

	def generate(self):
		if len(self.arguments) < 2:
			print(f"{Fore.RED}Error:{Style.RESET_ALL} Command 'generate' requires at least 1 parameters")
			print("Usage:")
			print("generate <name>")
			return
		
		result = generate(self.arguments[1], CLI)

	def unknown_command(self):
		print(f"{Fore.RED}Error:{Style.RESET_ALL} Unknown Command")

	def version_display(self):
		print(r""" /$$		/$$															   
| $$$	 /$$$															 
| $$$$	/$$$$  /$$$$$$	 /$$$$$$   /$$$$$$$ /$$   /$$  /$$$$$$	/$$   /$$
| $$ $$/$$ $$ /$$__  $$ /$$__  $$ /$$_____/| $$  | $$ /$$__  $$| $$  | $$
| $$  $$$| $$| $$$$$$$$| $$  \__/| $$	   | $$  | $$| $$  \__/| $$  | $$
| $$\  $ | $$| $$_____/| $$		 | $$	   | $$  | $$| $$	   | $$  | $$
| $$ \/  | $$|	$$$$$$$| $$		 |	$$$$$$$|  $$$$$$/| $$	   |  $$$$$$$
|__/	 |__/ \_______/|__/		  \_______/ \______/ |__/		\____  $$
																/$$  | $$
															   |  $$$$$$/
																\______/ """)
		print(version)



from colorama import Fore, Style
from json import loads, dumps
from time import time
import os
import inspect
import importlib.util

from sqlalchemy import Boolean, Float, Integer, String
from sqlalchemy.sql.schema import Table

def gen_log(message):
	print(f"{Fore.CYAN}[GENERATOR]{Style.RESET_ALL} {message}")

def get_model_from_path(path: str) -> Table:
	module_name = os.path.splitext(os.path.basename(path))[0]
	spec = importlib.util.spec_from_file_location(module_name, path)
	module = importlib.util.module_from_spec(spec)
	spec.loader.exec_module(module)
	model = None
	for name, obj in inspect.getmembers(module, inspect.isclass):
		if hasattr(obj, '__table__'):
			model = obj.__table__
			break
	if not type(model) == Table:
		raise Exception("This is impossible... wth happend?")
	return model

def generate_user(cli):
	table_name = input("What should be the table's name: ")
	username_field = input("What should the username field be called: ")
	password_field = input("What should the password field be called: ") 
	isHashed = input("Is the password hashed? (y/n): ")
	isEmail = input("Is there an email field? (y/n): ")
	email_field = ""
	email_command = ""
	if isEmail.lower().strip() == "y":
		email_field = input("What should the email field be called: ")
		email_command = f"{email_field} = Column(String(318))"

	c = cli([None, "create", "model", table_name])
	c.execute()
	if isHashed.lower().strip() == "n":
		with open(f"src/cargo/{table_name}Model.py", "w") as f:
			f.write(f"""#Generated by libmercury at {int(time())}
from libmercury.db import Column, Integer, Base, String

class {table_name}(Base):
	__tablename__ = "{table_name}"
	id = Column(Integer, primary_key=True)
	{username_field} = Column(String(20))
	{password_field} = Column(String(50))
	{email_command}
""")	
	elif isHashed.lower().strip() == "y":
		salt_field = input("What should the salt be called: ")
		with open(f"src/cargo/{table_name}Model.py", "w") as f:
			f.write(f"""#Generated by libmercury at {int(time())} 
from libmercury.db import Column, Integer, Base, String
from libmercury.security import generate_salt, hash_password, verify_password
import base64

class {table_name}(Base):
	__tablename__ = "{table_name}"
	id = Column(Integer, primary_key=True)
	{username_field} = Column(String(20))
	{password_field} = Column(String(50))
	{salt_field} = Column(String, nullable=False)
	{email_command}

	def set_password(self, password: str):
		self.{salt_field} = base64.b64encode(generate_salt()).decode('utf-8')
		self.{password_field} = hash_password(password, base64.b64decode(self.salt))

	def check_password(self, password: str) -> bool:
		return verify_password(self.{password_field}, password, base64.b64decode(self.{salt_field}))""")

	gen_log("Successfully created model")

def generate_login(cli):
	controller_name = input("What should be the name of the controller: ")
	controller_route = input("What should be the route of the controller: ")
	controller_success_redirect = input("What should be the success redirect of the controller: ")

	validator_name = input("What should be the name of the validator: ")

	model_name = input("What is the name of the model to sign into(leave blank to make a new one): ")
	if model_name == "":
		generate_user(cli)
		model_name = input("What is the name of the model to sign into(leave blank to make a new one): ")

	model_password_field = input("What is the name of the model's password field: ")
	model_username_field = input("What is the name of the model's username field: ")

	isHashed = input("Is the password hashed? (y/n): ")
	password_checker = f"not query(user, {model_username_field}={model_username_field}).first().{model_password_field} == {model_password_field}"

	if isHashed.lower().strip() == "y":
		validate_function = input("What is the name of the validator function(check_password if you used this for the model): ") 
		password_checker = f"not query(user, {model_username_field}={model_username_field}).first().{validate_function}({model_password_field})"

	jwt_name = input("What should be the name of the JWT: ")
	c = cli([None, "create", "jwt", jwt_name])
	c.execute()

	c = cli([None, "create", "controller", controller_name])
	c.execute()

	c = cli([None, "create", "validator", validator_name])
	c.execute()    
	
	with open(f"src/validators/{validator_name}Validator.py", "w") as f:
		f.write(f"""#Generated by libmercury at {int(time())}
from libmercury import Validator 
class {validator_name}Validator:
	{model_username_field} = Validator.String(max=20, min=5)
	{model_password_field} = Validator.String(max=100, min=8)""")
	
	with open(f"src/controllers/{controller_name}Controller.py", "w") as f:
		f.write(f"""#Generated by libmercury at {int(time())}
from libmercury import POSTRoute, Request, Response, redirect, useValidator, query, exists
from libmercury.utils import expires_in
from src.cargo.{model_name}Model import {model_name} 
from src.validators.{validator_name}Validator import {validator_name}Validator 
from src.security.{jwt_name}Jwt import {jwt_name}Jwt

class {controller_name}Controller:
	@staticmethod
	@POSTRoute("{controller_route}")
	@useValidator({validator_name}Validator)
	def login(request: Request) -> Response:
		if {jwt_name}Jwt._verify(request.cookies.get("token")):
			return redirect("{controller_success_redirect}")

		#Check to see if the account exists
		{model_username_field} = request.form["{model_username_field}"]
		if not exists(user, {model_username_field}={model_username_field}):
			#In production, just redirect to where the request will come from, do not rely on the referer header.
			referer = request.headers.get("Referer")
			if referer:
				return redirect(f"{{referer}}?error=Account doesn't exist")
			else:
				return redirect("/login?error=Account doesn't exist") 

		#Check to see if the password is correct
		{model_password_field} = request.form["{model_password_field}"]
		if {password_checker}:
			#In production, just redirect to where the request will come from, do not rely on the referer header.
			referer = request.headers.get("Referer")
			if referer:
				return redirect(f"{{referer}}?error=Wrong password")
			else:
				return redirect("/login?error=Wrong password")

		response = redirect("{controller_success_redirect}")
		response.set_cookie("token", {jwt_name}Jwt._makeJwt({{"username": username, "exp": expires_in(2592000)}}))
		return response 
""")
	gen_log("Successfully created login controller")

def generate_register(cli):
	controller_name = input("What should be the name of the controller: ")
	controller_route = input("What should be the route of the controller: ")
	controller_success_redirect = input("What should be the success redirect of the controller: ")

	validator_name = input("What should be the name of the validator: ")

	model_name = input("What is the name of the model to register for(leave blank to make a new one): ")
	if model_name == "":
		generate_user(cli)
		model_name = input("What is the name of the model to register for(leave blank to make a new one): ")

	jwt_checker = input("Provide the name of your JWT to make sure that users are not signed in, leave blank to make a new one: ")
	if jwt_checker == "":
		jwt_name = input("What should be the name of the JWT: ")
		c = cli([None, "create", "jwt", jwt_name])
		c.execute()

	username_field = input("What is the name of the model's username field: ")
	isHashed = input("Is the password hashed? (y/n): ")
	password_field = input("What is the name of the model's password field: ")
	set_passsword = ""
	if isHashed.lower().strip() == "y":
		set_password_field = input("What is the name of the model's password setter field: ")
		set_passsword = f"new_user.{set_password_field}({password_field})"
	salt_field = input("What is the name of the model's salt field: ")
	signed_in_redirect = input("Where should the controller redirect if the user is already signed in: ")
	primary_key = input("What is the name of the model's primary key: ")

	model = get_model_from_path(f"src/cargo/{model_name}Model.py")	
	if model == None:
		print(f"{Fore.RED}Error:{Style.RESET_ALL} Could not find model in src/cargo/{model_name}Model.py")
		return

	table = {}
	for col in model.columns:
		table[col.name] = col.type

	rules_python = []
	for key, value in table.items():
		if key == salt_field or key == primary_key:
			continue
		translated = ""
		if hasattr(value, "length") and value.length:
			value.length = f"max={value.length}"
		else:
			value.length = ""
		if type(value) == String:
			translated = f"Validator.String({value.length})"
		elif type(value) == Integer:
			translated = f"Validator.Integer({value.length})"
		elif type(value) == Boolean:
			translated = f"Validator.Boolean({value.length})"
		elif type(value) == Float:
			translated = f"Validator.Float({value.length})"
		else:
			print(f"[WARNING] Unhandled type for {key}, please add this to the validator and controller manually")
		rules_python.append(f"{key} = {translated}")
	rules_python = "\n\t".join(rules_python)

	c = cli([None, "create", "validator", validator_name])
	c.execute()  

	c = cli([None, "create", "controller", controller_name])
	c.execute()

	with open(f"src/validators/{validator_name}Validator.py", "w") as f:
		f.write(f"""#Generated by libmercury at {int(time())}
from libmercury import Validator
class {validator_name}Validator:
	{rules_python}""")

	lines = []
	user = f"new_user = {model_name}("

	for col in list(table.keys()):
		if col == primary_key or col == salt_field:
			continue
		lines.append(f'{col} = request.form["{col}"]')
		user = f"{user}{col}={col}," 
	user = user+")"
	lines = '\n\t\t'.join(lines)

	with open(f"src/controllers/{controller_name}Controller.py", "w") as f:
		f.write(f"""#Generated by libmercury at {int(time())}
from libmercury import useValidator, POSTRoute, redirect, Request, Response, exists
from src.validators.{validator_name}Validator import {validator_name}Validator 
from src.cargo.{model_name}Model import {model_name} 
from src.cargo.connection import Connection
from src.security.{jwt_checker}Jwt import {jwt_checker}Jwt
class RegisterController:
	@staticmethod
	@POSTRoute("{controller_route}")
	@useValidator(RegisterValidator)
	def {controller_name}(request: Request) -> Response:
		if {jwt_checker}Jwt._verify(request.cookies.get("token")):
			return redirect("{signed_in_redirect}")

		#Check to see if an account exists
		{username_field} = request.form["username"]
		if exists(user, {username_field}={username_field}):
			#In production, just redirect to where the request will come from, do not rely on the referer header.
			referer = request.headers.get("referer")
			if referer:
				return redirect(f"{{referer}}?error=Account already exists")
			else:
				return redirect("{controller_route}?error=Account already exists") 

		#Create the account
		{lines}
		{user}
		{set_passsword}
		Connection.Session.add_all([new_user,])
		Connection.Session.commit()

		#Redirect to the login page
		return redirect("{controller_success_redirect}")""")

def generate_password_reset(cli):
	pass

def generate_crud(cli):
	model_name = input("What is the name of the model: ")
	primary_key = input("What is the name of the model's primary key: ")
	controller_name = input("What should be the name of the controller: ")
	controller_route = input("What should be the route of the controller: ")
	validator_name = input("What should be the name of the validator: ")
	c = cli([None, "create", "controller", controller_name])
	c.execute()
	
	c = cli([None, "create", "validator", validator_name])
	c.execute()

	#Get the model information
	model = get_model_from_path(f"src/cargo/{model_name}Model.py")
	table = {}
	for col in model.columns:
		if not col.name == primary_key:
			table[col.name] = col.type

	#Write to our validator:
	rules_python = []
	for key, value in table.items():
		if key == "id":
			continue
		translated = ""
		if hasattr(value, "length") and value.length:
			value.length = f"max={value.length}"
		else:
			value.length = ""
		if type(value) == String:
			translated = f"Validator.String({value.length})"
		elif type(value) == Integer:
			translated = f"Validator.Integer({value.length})"
		elif type(value) == Boolean:
			translated = f"Validator.Boolean({value.length})"
		elif type(value) == Float:
			translated = f"Validator.Float({value.length})"
		else:
			print(f"{Fore.YELLOW}[WARNING]{Style.RESET_ALL} Unhandled type for {key}, please add this to the validator and controller manually")
		rules_python.append(f"{key} = {translated}")

	rules_python = "\n\t".join(rules_python)
	with open(f"src/validators/{validator_name}Validator.py", "w") as f:
		f.write(f"""#Generated by libmercury at {int(time())}
from libmercury import Validator
class {validator_name}Validator:
	{rules_python}""")
	
	#Write the controller
	with open(f"src/controllers/{controller_name}Controller.py", "w") as f:
		f.write(f"""#Generated by libmercury at {int(time())}
from json import dumps
from libmercury import DELETERoute, GETRoute, PATCHRoute, POSTRoute, Request, Response, find_or_404, json_to_object, object_to_json, update_object, useValidator
from src.cargo.{model_name}Model import {model_name} 
from src.validators.{validator_name}Validator import {validator_name}Validator
from src.cargo.connection import Connection

class {controller_name}Controller:
	@staticmethod
	@POSTRoute("{controller_route}")
	@useValidator({validator_name}Validator)
	def create(request: Request) -> Response:
		jsonobject = json_to_object(request.json, post)
		if type(jsonobject) == dict:
			#Return the error message
			return Response(dumps(jsonobject), content_type="application/json", status=400)
		Connection.Session.add_all([jsonobject])
		Connection.Session.commit()
		return Response(dumps({{"status":"success"}}))

	@staticmethod
	@GETRoute("{controller_route}/{{id:int}}")
	def read(request: Request, id: int) -> Response:
		find = find_or_404(post, "json", {primary_key}=id)
		if type(find) == Response:
			return find
		else:
			return Response(object_to_json(find_or_404(post, "json", id=id))) 

	@staticmethod
	@PATCHRoute("{controller_route}/{{id:int}}")
	def update(request: Request, id: int) -> Response:
		find = find_or_404(post, {primary_key}=id)
		if type(find) == Response:
			return find 
		return update_object(find, request.json)

	@staticmethod
	@DELETERoute("{controller_route}/{{id:int}}")
	def delete(request: Request, id: int) -> Response:
		find = find_or_404(post, {primary_key}=id)
		if type(find) == Response:
			return find 
		find.delete()
		return Response(dumps({{"status":"success"}}))""")
	
def generate(name, cli):
	generatable = {
		"model:user": generate_user, #Generates a model User
		"control:login": generate_login, #Generates a login controller with JWT's and validation
		"control:register": generate_register, #Generates a signup controller with validation
		"control:password_reset": generate_password_reset, #Generates a password reset controller
		"control:crud": generate_crud, #Generates a CRUD controller for a model
	}
	function = generatable.get(name)
	if not function:
		print(f"{Fore.RED}Error:{Style.RESET_ALL} Function not found")
	else:
		gen_log("Starting generation task")
		function(cli)



from werkzeug import Request
from werkzeug import Response
from werkzeug.utils import redirect
from .route_management import *
from .validation import Validator
from .validation.validate import validate
from .utils import *



import json
from werkzeug import Response
from werkzeug.utils import redirect
from sqlalchemy.exc import IntegrityError

def get_connection():
	from src.cargo.connection import Connection
	return Connection

def query(model, **kwargs):
	result = get_connection().Session.query(model).filter_by(**kwargs)
	return result

def exists(model, **kwargs):
	result = query(model, **kwargs).first()
	if result == None:
		return False
	return True

def reset_and_redirect(redirect_url, request):
    """Redirects to the specified URL and resets all cookies."""
    # Create a redirect response
    rsp = redirect(redirect_url)
    
    # Clear all cookies
    for cookie in request.cookies:
        rsp.delete_cookie(cookie)
    
    return rsp

def find_or_404(model, response_format="html", **kwargs):
	result = query(model, **kwargs).first()

	if result is None:
		if response_format == "json":
			error_message = {
				"error": "404 Not Found",
				"details": f"{model.__name__} not found with {kwargs}"
			}
			response_body = json.dumps(error_message)
			return Response(response_body, status=404, mimetype='application/json')
		else:  # Default to HTML
			error_message = f"<h1>404 Not Found</h1><p>{model.__name__} not found with {kwargs}</p>"
			return Response(error_message, status=404, mimetype='text/html')
	
	return result

def paginate(model, page=1, per_page=10, **kwargs):
	"""Paginate the results of a query on a SQLAlchemy model, returning a query object."""
	query_result = query(model, **kwargs)  # Assuming `query` is a function that returns a query object
	paginated_query = query_result.offset((page - 1) * per_page).limit(per_page)
	
	return paginated_query

def expires_in(seconds: int):
	import time
	return int(time.time())+seconds

def object_to_json(model, exclude=None):
	"""Converts a SQLAlchemy model instance to a JSON string, with optional exclusion of specified fields."""
	if exclude is None:
		exclude = []

	def model_to_dict(obj):
		"""Converts a SQLAlchemy model instance to a dictionary, excluding specified fields."""
		return {
			c.name: getattr(obj, c.name)
			for c in obj.__table__.columns
			if c.name not in exclude
		}

	model_dict = model_to_dict(model)
	return json.dumps(model_dict)

def json_to_object(json_dict, model_class):
	"""Converts a JSON string to a SQLAlchemy model instance, excluding the primary key.

	Args:
		json_str (str): The JSON string representing the model data.
		model_class (SQLAlchemy Model): The SQLAlchemy model class to instantiate.

	Returns:
		object or dict: The SQLAlchemy model instance if successful, or a dictionary with an error message.
	"""
	try:
		# Parse the JSON string
		data = json_dict
		# Get the primary key column name
		primary_key_column = model_class.__table__.primary_key.columns.keys()[0]
		
		# Ensure the primary key is not included in the JSON
		if primary_key_column in data:
			raise ValueError(f"Primary key '{primary_key_column}' should not be provided in the JSON.")
		
		# Instantiate the model using the remaining data
		instance = model_class(**data)
		return instance

	except json.JSONDecodeError as e:
		# Handle JSON parsing errors
		return {"error": "Invalid JSON format", "details": str(e)}

	except ValueError as e:
		# Handle missing or invalid primary key
		return {"error": "Validation error", "details": str(e)}

	except TypeError as e:
		# Handle errors where JSON keys don't match model attributes
		return {"error": "Type error", "details": str(e)}

	except IntegrityError as e:
		# Handle SQLAlchemy integrity errors
		return {"error": "Integrity error", "details": str(e)}

	except Exception as e:
		# Handle any other exceptions
		return {"error": "Unexpected error", "details": str(e)}

def update_object(object, updates):
	session = get_connection().Session
	"""Update an SQLAlchemy object with the provided dictionary of changes.
	
	Args:
		object (SQLAlchemy model instance): The object to update.
		updates (dict): Dictionary of field updates.

	Returns:
		werkzeug.wrappers.Response: A response object indicating success or error.
	"""
	# Get the primary key and unique columns
	primary_key_column = object.__table__.primary_key.columns.keys()[0]
	unique_columns = [c.name for c in object.__table__.columns if c.unique]

	# Check if ID or unique fields are being changed
	for key in updates:
		if key == primary_key_column or key in unique_columns:
			error_message = {
				"error": "Invalid update attempt",
				"details": f"Cannot change '{key}' field."
			}
			return Response(
				json.dumps(error_message),
				status=400,
				mimetype='application/json'
			)

	# Apply updates to the object
	for key, value in updates.items():
		setattr(object, key, value)

	# Commit the changes
	try:
		session.commit()
		success_message = {
			"success": "Update successful",
			"details": f"Updated {object.__class__.__name__} with ID {getattr(object, primary_key_column)}."
		}
		return Response(
			json.dumps(success_message),
			status=200,
			mimetype='application/json'
		)
	except IntegrityError as e:
		session.rollback()
		error_message = {
			"error": "Database error",
			"details": str(e.orig)	# You might need to adjust based on your DB errors
		}
		return Response(
			json.dumps(error_message),
			status=400,
			mimetype='application/json'
		)
	except Exception as e:
		session.rollback()
		error_message = {
			"error": "Unexpected error",
			"details": str(e)
		}
		return Response(
			json.dumps(error_message),
			status=500,
			mimetype='application/json'
		)




from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15
import json
import hmac
import hashlib
import time
import base64
class JWT:
	def __init__(self, token):
		self.token = token
		self.header = {} 
		self.payload = {} 
		self.signature = "" 
		if token != "" and token != None:
			self._parse_token()

	def _parse_token(self):
		parts = self.token.split('.')
		if len(parts) != 3:
			raise ValueError("Invalid JWT token format")

		self.header = self._base64_url_decode(parts[0])
		self.payload = self._base64_url_decode(parts[1])
		self.signature = parts[2]

	def _base64_url_decode(self, input_str):
		# Add padding if needed
		input_str += '=' * (4 - len(input_str) % 4)
		decoded_bytes = base64.urlsafe_b64decode(input_str)
		return json.loads(decoded_bytes)

	def _base64_url_encode(self, data):
		encoded = base64.urlsafe_b64encode(data).decode('utf-8')
		return encoded.rstrip('=')

	def get_header(self):
		return self.header

	def get_payload(self):
		return self.payload

	def get_signature(self):
		return self.signature

	def verify_signature(self, pem_file_path):
		if self.token == None or self.token == "":
			return False

		#Check to see if the token is valid
		try:
			if self.payload.get("exp") and int(time.time()) > int(self.payload.get("exp")):
				return False
		except ValueError:
			return False

		with open(pem_file_path, 'rb') as pem_file:
			key_data = pem_file.read()

		header_encoded, payload_encoded, signature_encoded = self.token.split('.')
		
		if self.header['alg'] == 'HS256':
			expected_signature = hmac.new(
				key_data,
				f'{header_encoded}.{payload_encoded}'.encode(),
				hashlib.sha256
			).digest()
			
			# Decode the JWT signature
			decoded_signature = base64.urlsafe_b64decode(self.signature + '=' * (4 - len(self.signature) % 4))

			return hmac.compare_digest(decoded_signature, expected_signature)
		
		elif self.header['alg'] == 'RS256':
			message = f'{header_encoded}.{payload_encoded}'.encode()
			decoded_signature = base64.urlsafe_b64decode(self.signature + '=' * (4 - len(self.signature) % 4))
			rsa_key = RSA.import_key(key_data)
			h = SHA256.new(message)
			try:
				pkcs1_15.new(rsa_key).verify(h, decoded_signature)
				return True
			except (ValueError, TypeError):
				return False

		else:
			return False

	def to_string(self):
		header_encoded = self._base64_url_encode(json.dumps(self.header).encode())
		payload_encoded = self._base64_url_encode(json.dumps(self.payload).encode())
		return f"{header_encoded}.{payload_encoded}.{self.signature}"

	def sign(self, private_key_path, key_type):
		if key_type.upper() == "RSA":
			key_type = "RS256"
		else:
			key_type = "HS256"

		if self.header == {}:
			self.header = {"alg": key_type, "typ": "JWT"}

		with open(private_key_path, 'rb') as pem_file:
			key_data = pem_file.read()

		header_encoded = self._base64_url_encode(json.dumps(self.header).encode())
		payload_encoded = self._base64_url_encode(json.dumps(self.payload).encode())
		message = f"{header_encoded}.{payload_encoded}".encode()

		if key_type == 'HS256':
			new_signature = hmac.new(key_data, message, hashlib.sha256).digest()
		elif key_type == 'RS256':
			rsa_key = RSA.import_key(key_data)
			h = SHA256.new(message)
			new_signature = pkcs1_15.new(rsa_key).sign(h)
		else:
			raise ValueError(f"Unsupported algorithm: {key_type}")
		self.signature = self._base64_url_encode(new_signature)
		return self.to_string()



import os
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from colorama import Fore, Style

def save_key_to_file(path, filename, key_data):
	os.makedirs(path, exist_ok=True)
	with open(os.path.join(path, filename), 'wb') as file:
		file.write(key_data)

def generate_hmac_key(name, path):
	secret_key = get_random_bytes(32)  # 256-bit key
	save_key_to_file(path, f'{name}Hmac_secret.key', secret_key)
	print(f"{Fore.MAGENTA}[KEYGEN]{Style.RESET_ALL} HMAC secret key saved to {os.path.join(path, f'{name}Hmac_secret.key')}")

def generate_rsa_keys(name, path):
	key = RSA.generate(2048)
	private_key = key.export_key()
	public_key = key.publickey().export_key()
	save_key_to_file(path, f'{name}Private_key.pem', private_key)
	save_key_to_file(path, f'{name}Public_key.pem', public_key)
	print(f"{Fore.MAGENTA}[KEYGEN]{Style.RESET_ALL} RSA private key saved to {os.path.join(path, f'{name}Public_key.pem')}")
	print(f"{Fore.MAGENTA}[KEYGEN]{Style.RESET_ALL} RSA public key saved to {os.path.join(path, f'{name}Private_key.pem')}")

def main(name):
	path = 'src/.vault'
	key_type = input("What kind of key do you want to generate? (HMAC/RSA): ").strip().upper()

	if key_type == 'HMAC':
		generate_hmac_key(name, path)
		return "HMAC"
	elif key_type == 'RSA':
		generate_rsa_keys(name, path)
		return "RSA"
	else:
		print("Invalid key type. Please choose either 'HMAC' or 'RSA'.")



from colorama import Fore, Style
from sqlalchemy import Column, create_engine, MetaData, Table
from sqlalchemy.exc import SQLAlchemyError
import importlib.util
import os
import inspect

class MigrationSystem:
	def __init__(self, db_connection_path: str, model_paths: list) -> None:
		self.db_connection_path = db_connection_path
		self.model_paths = model_paths

	def _extract_db_url_from_connection(self, file_path: str) -> None:
		# Load the module from the file path
		module_name = os.path.splitext(os.path.basename(file_path))[0]
		spec = importlib.util.spec_from_file_location(module_name, file_path)
		module = importlib.util.module_from_spec(spec)
		spec.loader.exec_module(module)
		
		# Extract the database URL from the `Connection` object
		if hasattr(module, 'Connection'):
			connection = module.Connection
			# Extracting the database URL from the connection object
			# This depends on the actual implementation of `connection` in `libmercury.db`
			if hasattr(connection, 'Engine'):
				db_url = connection.Engine.url
				return str(db_url)
			else:
				raise AttributeError("The 'Connection' object does not have an 'engine' or 'url' attribute.")
		else:
			raise AttributeError("The module does not have a 'Connection' object.")


	def _generate_file(self, autogenerate_table, message) -> None:
		python_files = []
	
		for file in os.listdir("src/cargo/migrations"):
			if file.endswith('.py') and os.path.isfile(os.path.join("src/cargo/migrations", file)):
				try:
					python_files.append(int(file[:-3]))
				except ValueError:
					pass
	
		name = max(python_files) + 1 if python_files else 1
	
		upgrade_commands = []
		downgrade_commands = []
	
		# Separate tables with and without foreign key dependencies
		fk_tables = []
		no_fk_tables = []
	
		for new_table in autogenerate_table.get("new_tables", []):
			if any(col_info.get("foreign_keys") for col_info in new_table["columns"].values()):
				fk_tables.append(new_table)
			else:
				no_fk_tables.append(new_table)
	
		# First stage: Handle tables without foreign keys
		for new_table in no_fk_tables:
			table_name = new_table["name"]
			columns = []
	
			for col_name, col_info in new_table["columns"].items():
				col_type = col_info["type"]
				col_nullable = col_info["nullable"]
				col_primary_key = col_info["primary_key"]
	
				column_def = f"Column('{col_name}', {col_type}"
				if col_primary_key:
					column_def += f", primary_key={col_primary_key}"
				column_def += f", nullable={col_nullable}"
				column_def += ")"
	
				columns.append(column_def)
	
			columns_str = ", \n\t\t".join(columns)
			upgrade_commands.append(f"wrapper.create_table('{table_name}', [\n\t\t{columns_str}\n\t])")
			downgrade_commands.append(f"wrapper.delete_table('{table_name}')")
	
		# Second stage: Handle tables with foreign keys
		for new_table in fk_tables:
			table_name = new_table["name"]
			columns = []
	
			for col_name, col_info in new_table["columns"].items():
				col_type = col_info["type"]
				col_nullable = col_info["nullable"]
				col_primary_key = col_info["primary_key"]
				col_foreign_keys = col_info.get("foreign_keys")
	
				column_def = f"Column('{col_name}', {col_type}"
				if col_primary_key:
					column_def += f", primary_key={col_primary_key}"
				if col_foreign_keys:
					column_def += f", {col_foreign_keys[0]}"
				column_def += f", nullable={col_nullable}"
				column_def += ")"
	
				columns.append(column_def)
	
			columns_str = ", \n\t\t".join(columns)
			upgrade_commands.append(f"wrapper.create_table('{table_name}', [\n\t\t{columns_str}\n\t])")
			downgrade_commands.append(f"wrapper.delete_table('{table_name}')")
	
		# Handle removed columns (drop columns first)
		for removed_column in autogenerate_table.get("removed_columns", []):
			table_name = removed_column["table"]
			column_name = removed_column["name"]
	
			upgrade_commands.append(f"wrapper.drop_column('{table_name}', '{column_name}')")
			downgrade_commands.append(f"wrapper.add_column('{table_name}', {column_name})")  # Assuming re-adding logic
	
		# Handle new columns (add columns after dropping)
		for new_column in autogenerate_table.get("new_columns", []):
			table_name = new_column["table"]
			column_name = new_column["name"]
			column_type = new_column["type"]
			nullable = new_column["nullable"]
			primary_key = new_column["primary_key"]
			foreign_keys = new_column.get("foreign_keys")
	
			col_def = f"Column('{column_name}', {column_type}"
			if primary_key:
				col_def += f", primary_key={primary_key}"
			if foreign_keys:
				col_def += f", ForeignKey('{foreign_keys[0]}')"
			col_def += f", nullable={nullable}"
			col_def += ")"
	
			upgrade_commands.append(f"wrapper.add_column('{table_name}', {col_def})")
			downgrade_commands.append(f"wrapper.drop_column('{table_name}', '{column_name}')")
	
		# Write to file
		with open(f"src/cargo/migrations/{name}.py", "w") as f:
			f.write("from libmercury.db import MigrationWrapper, Column, INTEGER, VARCHAR, ForeignKey\n\n")
			f.write(f"_version = '{name}'\n")
			f.write(f"_prev_version = '{int(name) - 1}'\n\n")
			f.write(f"_commit_message = '{message}'\n\n")
			f.write("def upgrade(url):\n")
			f.write("\twrapper = MigrationWrapper(url)\n")
			for cmd in upgrade_commands:
				f.write(f"\t{cmd}\n")
	
			f.write("\n\ndef downgrade(url):\n")
			f.write("\twrapper = MigrationWrapper(url)\n")
			for cmd in downgrade_commands:
				f.write(f"\t{cmd}\n")
	
		print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Generated file: src/cargo/migrations/{name}.py")

	def _create_migration(self, message) -> None:
		# Step 0: Load ORM models
		print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Loading ORM models")
		orm_models = self.load_orm_models(self.model_paths)
		
		# Step 2: Get the database schema
		print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Loading database schema")
		db_metadata = self.get_database_schema(self._extract_db_url_from_connection(self.db_connection_path))
		
		# Step 3: Compare schemas
		print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Finding discrepancies...")
		migration_response = self.compare_schemas(orm_models, db_metadata)
		discrepancies = migration_response[0]
		autogenerate_table = migration_response[1]
		
		# Output discrepancies
		print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Loading discrepancies")
		if discrepancies:
			for discrepancy in discrepancies:
				print(discrepancy)
			self._generate_file(autogenerate_table, message)
		else:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} ORM models perfectly match the database schema, no migrations are required")

	def load_orm_models(self, model_paths: list) -> list:
		models = []
		for path in model_paths:
			module_name = os.path.splitext(os.path.basename(path))[0]
			spec = importlib.util.spec_from_file_location(module_name, path)
			module = importlib.util.module_from_spec(spec)
			spec.loader.exec_module(module)
			for name, obj in inspect.getmembers(module, inspect.isclass):
				if hasattr(obj, '__table__'):
					models.append(obj)
		return models

	def get_database_schema(self, engine_url: str) -> MetaData:
		engine = create_engine(engine_url)
		metadata = MetaData()
		metadata.reflect(bind=engine)
		return metadata

	def compare_schemas(self, orm_models: list, db_metadata: MetaData) -> tuple:
		discrepancies = []
		autogenerate_table = {"new_columns": [], "new_tables": [], "removed_columns": []}
		orm_tables = {}
	
		for model in orm_models:
			table = model.__table__
			orm_tables[table.name] = {
				col.name: {
					"type": col.type,
					"nullable": col.nullable,
					"primary_key": col.primary_key,
					"foreign_keys": list(col.foreign_keys)	# Store foreign key information
				}
				for col in table.columns
			}
	
		db_tables = {
			table.name: {
				col.name: {
					"type": col.type,
					"nullable": col.nullable,
					"primary_key": col.primary_key,
					"foreign_keys": list(col.foreign_keys)	# Store foreign key information
				}
				for col in table.columns
			}
			for table in db_metadata.sorted_tables
		}
	
		for table_name, orm_columns in orm_tables.items():
			if table_name not in db_tables:
				discrepancies.append(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Table '{table_name}' is missing in the database.")
				autogenerate_table["new_tables"].append({
					"name": table_name,
					"columns": orm_columns
				})
				continue
	
			db_columns = db_tables[table_name]
	
			for col_name, orm_col_data in orm_columns.items():
				if col_name not in db_columns:
					discrepancies.append(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Column '{col_name}' in table '{table_name}' is missing in the database.")
					autogenerate_table["new_columns"].append({
						"name": col_name,
						"type": orm_col_data["type"],
						"table": table_name,
						"nullable": orm_col_data["nullable"],
						"primary_key": orm_col_data["primary_key"],
						"foreign_keys": [fk.target_fullname for fk in orm_col_data["foreign_keys"]]
					})
				else:
					db_col_data = db_columns[col_name]
					if (str(orm_col_data["type"]) != str(db_col_data["type"]) or
						orm_col_data["nullable"] != db_col_data["nullable"] or
						orm_col_data["primary_key"] != db_col_data["primary_key"]):
						discrepancies.append(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Column '{col_name}' in table '{table_name}' has type or attribute mismatch.")
						autogenerate_table["removed_columns"].append({
							"table": table_name,
							"name": col_name
						})
						autogenerate_table["new_columns"].append({
							"name": col_name,
							"type": orm_col_data["type"],
							"table": table_name,
							"nullable": orm_col_data["nullable"],
							"primary_key": orm_col_data["primary_key"],
							"foreign_keys": [fk.target_fullname for fk in orm_col_data["foreign_keys"]]
						})
	
			for col_name in db_columns:
				if col_name not in orm_columns:
					discrepancies.append(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Column '{col_name}' in table '{table_name}' is extra in the database.")
					autogenerate_table["removed_columns"].append({
						"table": table_name,
						"name": col_name
					})
	
		return discrepancies, autogenerate_table

class MigrationWrapper:
	def __init__(self, connection_string: str) -> None:
		self.engine = create_engine(connection_string)
		self.metadata = MetaData(bind=self.engine)

	def create_table(self, table_name: str, columns: list) -> None:
		"""
		Create a new table with specified columns.
		
		:param table_name: Name of the table to create
		:param columns: List of Column definitions
		"""
		self.metadata.reflect(bind=self.engine)
		try:
			table = Table(table_name, self.metadata, *columns)
			table.create(self.engine)
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Table '{table_name}' created successfully.")
		except SQLAlchemyError as e:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Error creating table: {e}")

	def delete_table(self, table_name: str) -> None:
		"""
		Delete an existing table.
		
		:param table_name: Name of the table to delete
		"""
		try:
			table = Table(table_name, self.metadata, autoload_with=self.engine)
			table.drop(self.engine)
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Table '{table_name}' deleted successfully.")
		except SQLAlchemyError as e:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Error deleting table: {e}")

	def add_column(self, table_name: str, column: Column) -> None:
		"""
		Add a new column to an existing table.
	
		:param table_name: Name of the table
		:param column: Column definition
		"""
		try:
			table = Table(table_name, self.metadata, autoload_with=self.engine)
			with self.engine.connect() as conn:
				# Manually construct the column definition with type, nullable, and default value
				column_sql = f"{column.name} {column.type.compile(self.engine.dialect)}"
			
				if not column.nullable:
					column_sql += " NOT NULL"
				else:
					column_sql += " NULL"
			
				if column.default is not None:
					# Extract the default value, accounting for SQL expressions or callable defaults
					if callable(column.default.arg):
						default_value = column.default.arg()
					else:
						default_value = column.default.arg
					column_sql += f" DEFAULT {default_value}"
			
				conn.execute(f'ALTER TABLE {table_name} ADD COLUMN {column_sql}')
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Column '{column.name}' added to table '{table_name}'.")
		except SQLAlchemyError as e:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Error adding column: {e}")

	def drop_column(self, table_name: str, column_name: str) -> None:
		"""
		Drop an existing column from a table.
		
		:param table_name: Name of the table
		:param column_name: Name of the column to drop
		"""
		try:
			table = Table(table_name, self.metadata, autoload_with=self.engine)
			with self.engine.connect() as conn:
				conn.execute(f'ALTER TABLE {table_name} DROP COLUMN {column_name}')
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Column '{column_name}' dropped from table '{table_name}'.")
		except SQLAlchemyError as e:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Error dropping column: {e}")

	def modify_column(self, table_name: str, old_column_name: str, new_column: Column) -> None:
		"""
		Modify an existing column in a table.
		
		:param table_name: Name of the table
		:param old_column_name: Name of the column to modify
		:param new_column: New Column definition
		"""
		try:
			table = Table(table_name, self.metadata, autoload_with=self.engine)
			with self.engine.connect() as conn:
				conn.execute(f'ALTER TABLE {table_name} RENAME COLUMN {old_column_name} TO temp_{old_column_name}')
				conn.execute(f'ALTER TABLE {table_name} ADD COLUMN {new_column.compile(conn.dialect)}')
				conn.execute(f'UPDATE {table_name} SET {new_column.name} = temp_{old_column_name}')
				conn.execute(f'ALTER TABLE {table_name} DROP COLUMN temp_{old_column_name}')
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Column '{old_column_name}' modified to '{new_column.name}' in table '{table_name}'.")
		except SQLAlchemyError as e:
			print(f"{Fore.GREEN}[Migrator]{Style.RESET_ALL} Error modifying column: {e}")




from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

class connection:
	def __init__(self, *args, **kwargs) -> None:
		self.Engine = create_engine(*args, **kwargs)
		self._sessionmaker = sessionmaker(bind=self.Engine)
		self.Session = self._sessionmaker()



from werkzeug import Response

def validate(validator, error, data):
	class_vars = {}
	for name, value in validator.__dict__.items():
		if not name.startswith("__"):
			if getattr(value, "__exclude_from_validation", False):
				continue  # Skip this field if marked as excluded
			class_vars[name] = value

	class_fields = list(class_vars.keys())
	request_fields = list(data.keys())

	for field in class_fields:
		if field not in request_fields:
			if error:
				return error()
			rsp = Response(f"Error: Missing field '{field}'")
			rsp.status_code = 400
			return rsp
		else:
			value = data[field]
			field_validator = class_vars[field]
			if hasattr(field_validator,  "__nested_validator"):
				if type(value) != dict:
					rsp = Response(f"Error: type mismatch in field '{field}'")
					rsp.status_code = 400
					return rsp
				nested_result = validate(field_validator, error, value) 
				if nested_result:
					return nested_result
			elif not field_validator.validate(value):
				if error:
					return error()
				rsp = Response(f"Error: type mismatch in field '{field}'")
				rsp.status_code = 400
				return rsp



from .MigrationGenerator import MigrationWrapper
from .connection import connection
from sqlalchemy.ext.declarative import declarative_base
from libmercury.db import Column, Integer

b = declarative_base()
class mercury_version(b):
	__tablename__ = "mercury_version"
	id = Column(Integer(), primary_key=True)
	version = Column(Integer())

def create_mercury_table(url):
	wrapper = MigrationWrapper(url)
	wrapper.create_table("mercury_version", [
		Column("id", Integer, primary_key=True),
		Column("version", Integer)
	])

def find_mercury_table(url):
	conn = connection(url)
	base = declarative_base()
	base.metadata.reflect(conn.Engine)
	table = base.metadata.tables.get("mercury_version")
	return table

def get_version(url):
	conn = connection(url)
	m = conn.Session.query(mercury_version).first()
	return m.version

def create_version(url, version):
	conn = connection(url)
	m = mercury_version(version=version)
	conn.Session.add(m)
	conn.Session.commit()

def update_version(url, version):
	conn = connection(url)
	m = conn.Session.query(mercury_version).first()
	m.version = version 
	conn.Session.add(m)
	conn.Session.commit()



from functools import wraps
from jinja2 import Environment, FileSystemLoader
from .security import JWT
from werkzeug import Response

def useAuthorization(auth, **kwargs):
	def decorator(func):
		@wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, **kwargs)
		wrapper._auth = auth 
		wrapper._auth_cookie = kwargs.get("cookie")
		wrapper._error = kwargs.get("error")
		wrapper._negative_auth = kwargs.get("negative_auth")
		wrapper._jwt_validator = kwargs.get("validator")
		return wrapper
	return decorator

def useValidator(validator, **kwargs):
	def decorator(func):
		@wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, **kwargs)
		wrapper._validator = validator
		wrapper._error = kwargs.get("error")
		wrapper._mimetypes = kwargs.get("mimetypes")
		return wrapper
	return decorator

def useJWTValidator(validator, **kwargs):
	def decorator(func):
		@wraps(func)
		def wrapped_function(*args, **kwargs):
			from libmercury.validation import validate
			jwt = args[0]  # Assuming the JWT is the first argument
			error = kwargs.get('error')  # Get the error handler or message from kwargs
			func_result = func(*args, **kwargs)
			if func_result == False:
				return False
			validation_result = validate(validator, error, JWT(jwt).payload)
			if not validation_result:
				return True
			else:
				return False
		return wrapped_function
	return decorator

def route(method, url):
	def decorator(func):
		@wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, **kwargs)
		wrapper._route_method = method
		wrapper._route_url = url
		return wrapper
	return decorator
# HTTP method specific decorators
def GETRoute(url):
	return route('GET', url)

def POSTRoute(url):
	return route('POST', url)

def DELETERoute(url):
	return route('DELETE', url)

def PATCHRoute(url):
	return route('PATCH', url)

def PUTRoute(url):
	return route('PUT', url)

def OPTIONSRoute(url):
	return route('OPTIONS', url)

def HEADRoute(url):
	return route('HEAD', url)

def CONTROLRoute(url):
	return route('CONTROL', url) 

def TRACERoute(url):
	return route('TRACE', url)

class Route:
	def __init__(self, method, url, handler):
		self.method = method
		self.url = url
		self.handler = handler

	def __repr__(self):
		return f"Route(method={self.method}, url='{self.url}', handler={self.handler})"

def use_template(template_name: str, **kwargs):
	enviroment = Environment(loader=FileSystemLoader("src/templates"))
	template = enviroment.get_template(template_name)
	response = Response(template.render(kwargs))
	response.headers["Content-Type"] = "text/html"
	return response
